"""Provide methods to manipulate Bitwarden vault using the Bitwarden CLI

"""
from copy import deepcopy
import json
import logging
from subprocess import run


def status(session=b''):
    """Check status of vault

        Returns: Dict -
                 {serverUrl: <url>,
                 lastSync: date/time,
                 userEmail: <email>,
                 userId: userId,
                 status: <'locked', 'unlocked' or 'unauthenticated'>}
            False on error

    """
    res = run(["bw", "--session", session, "status"], capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return dict(json.loads(res.stdout.split(b'\n')[-1]))


def set_server(url="https://vault.bitwarden.com"):
    """Set vault URL

        Returns: True if successful or False on error

    """
    res = run(["bw", "config", "server", url], capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return True


def login(email, password, method=None, code=""):
    """Initial login to Bitwarden Vault.

        Args: email - string
              password - string
              method - int (0: Authenticator, 1: Email, 3: Yubikey)
              code - OTP code

        Returns: session (bytes) or False on error, Error message

    """
    cmd = ["bw", "login", "--raw", email, password]
    if method and code:
        cmd = ["bw", "login", "--raw", email, password, "--method", method, "--code", code]
    res = run(cmd, capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return (False, res.stderr)
    return res.stdout, None


def unlock(password):
    """Unlock vault

        Returns: session (bytes) or False on error, Error message

    """
    res = run(["bw", "unlock", "--raw", password], capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return (False, res.stderr)
    return res.stdout, None


def lock():
    """Lock vault

        Return: True on success, False with any errors

    """
    res = run(["bw", "lock"], capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return True


def logout():
    """Logout of vault

        Return: True on success, False with any errors

    """
    res = run(["bw", "logout"], capture_output=True, check=False)
    if not res.stderr:
        logging.error(res)
        return False
    return True


def get_orgs(session):
    """Return all organizations for the logged in user

        Return: Dict of org dicts {id: dict('object':'organization','id':id,'name':<name>...)}
                False on error

    """
    res = run(["bw", "--session", session, "list", "organizations"],
              capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return {i['id']: i for i in json.loads(res.stdout)}


class Item(dict):
    """Set some default attributes to all items

    """
    def __init__(self, /, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self['login']['url'] = self['login']['uris'][0]['uri']
        except (IndexError, KeyError):
            self.setdefault('login', {})
            self['login'].setdefault('url', "")
        self.setdefault('fields', [])
        if not any(i['name'] == 'autotype' for i in self.get('fields')):
            self['fields'].append({'name': 'autotype', 'value': "", 'type': 0})


def get_entries(session=b'', org_name=''):
    """Get all entries, folders, collections and orgs from vault

    Args: session: bytes
          org_name: name of organization. If given, only return items for that org
    1. since the URL is buried in:
        'login'->'uris'->[{match: xxx, uri: http...}, {match2: xxx, uri2: httpxxx}]
        copy the first uri to 'login'->'url' for ease of access later.
    2. Also adjust 'path' to be just the dirname, not including the 'name'
    3. Add the 'autotype' field so it can be edited if necessary

        Return: items (list of Items), folders, collections, orgs
                False on error

    """
    res = run(["bw", "--session", session, "list", "items"], capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    items = [Item(i) for i in json.loads(res.stdout)]
    folders = get_folders(session)
    collections = get_collections(session, org_name)
    orgs = get_orgs(session)
    return items, folders, collections, orgs


def sync(session=b''):
    """Sync web vault changes to local vault

        Return: True on success, False with any errors

    """
    res = run(["bw", "--session", session, "sync"], capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return True


def get_folders(session):
    """Return all folder names.

        Return: Dict of folder dicts {id: dict('object':folder,'id':id,'name':<name>)}
                False on error

    """
    res = run(["bw", "--session", session, "list", "folders"], capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return {i['id']: i for i in json.loads(res.stdout)}


def get_collections(session, org_id=''):
    """Return all collection names for user.

        Args: session - session id bytes
              org_id - organization id string.

        Return: Dict of collection dicts {id:
            dict('object':collection,'id':id,'organizationId:<org
                 id>,'externalId':<ext id>,'name':<name>)}

    """
    cmd = ["bw", "--session", session, "list", "collections"]
    if org_id:
        cmd.extend(["--organizationid", org_id])
    res = run(cmd, capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return {i['id']: i for i in json.loads(res.stdout)}


def add_entry(entry, session):
    """Add new entry to vault

    Args: entry - dict with at least these fields
    Returns: new item dict or False on error

    New item template:
    {
        "organizationId":null,
        "folderId":null,
        "type":1,
        "name":"Item name",
        "notes":"Some notes about this item.",
        "favorite":false,
        "fields":[],
        "login":null,
        "secureNote":null,
        "card":null,
        "identity":null}'

    """
    enc = run(["bw", "encode"], input=json.dumps(entry).encode(), capture_output=True, check=False)
    if not enc.stdout:
        logging.error(enc)
        return False
    res = run(["bw", "create",
               "--session", session,
               "item", enc.stdout],
              capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return json.loads(res.stdout)


def edit_entry(entry, session, update_coll='NO'):
    # pylint: disable=too-many-return-statements
    """Modify existing vault entry

        Args: entry - entry dict object
              session - session id
              update_coll - 'YES' if item collections have been modified.
                            'MOVE' if collections are added so item needs to be
                                move to an org.  Returns: updated entry object
                                (dict) on success, False on failure
                            'REMOVE' if collections are removed so item needs to
                                be moved from the org to personal vault.

    """
    item = deepcopy(entry)
    if update_coll == 'YES':
        enc = run(["bw", "encode"],
                  input=json.dumps(item['collectionIds']).encode(),
                  capture_output=True,
                  check=False)
        if not enc.stdout:
            logging.error(enc)
            return False
        res = run(["bw", "edit",
                   "--session", session,
                   "item-collections", item['id'], enc.stdout],
                  capture_output=True, check=False)
        if not res.stdout:
            logging.error(res)
            return False
    elif update_coll == 'MOVE':
        res = move_entry(entry, session)
        if res is False:
            return False
    elif update_coll == 'REMOVE':
        res = delete_entry(entry, session)
        if res is False:
            return False
        item['id'] = None
        item['collectionIds'] = []
        item['organizationId'] = None
        res = add_entry(item, session)
        if res is False:
            return False
        return res
    enc = run(["bw", "encode"], input=json.dumps(item).encode(), capture_output=True, check=False)
    if not enc.stdout:
        logging.error(enc)
        return False
    res = run(["bw", "edit",
               "--session", session,
               "item", item['id'], enc.stdout],
              capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return json.loads(res.stdout)


def move_entry(entry, session):
    """Move entry to an organization (if it currently belongs to personal vault)

        Assumes entry already has the organizationId and collectionIds updated.

        Args: entry - entry dict object
              session - session id
        Returns: updated entry object (dict) on success, False on failure

    """
    item = deepcopy(entry)
    enc = run(["bw", "encode"],
              input=json.dumps(item['collectionIds']).encode(),
              capture_output=True,
              check=False)
    if not enc.stdout:
        logging.error(enc)
        return False
    res = run(["bw", "move",
               "--session", session,
               item['id'],
               item['organizationId'].encode(),
               enc.stdout],
              capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return json.loads(res.stdout)


def delete_entry(entry, session):
    """Delete existing vault entry

        Args: entry - entry dict object
        Returns: entry object (dict) on success, False on failure

    """
    res = run(["bw", "delete",
               "--session", session,
               "item", entry['id']],
              capture_output=True, check=False)
    if res.returncode != 0:
        logging.error(res)
        return False
    return entry


def add_folder(folder, session):
    """Add folder

        Args: folder - string (name of folder)
              session - bytes

        Returns: Folder object or False on error

    """
    folder = {"name": folder}
    enc = run(["bw", "encode"], input=json.dumps(folder).encode(), capture_output=True, check=False)
    if not enc.stdout:
        logging.error(enc)
        return False
    res = run(["bw", "create",
               "--session", session,
               "folder", enc.stdout],
              capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return json.loads(res.stdout)


def delete_folder(folder, session):
    """Delete folder

        Args: folder - folder object (dict)
              session - bytes
        Returns: folder object (dict) on success, False on failure


    """
    res = run(["bw", "delete",
               "--session", session,
               "folder", folder['id']],
              capture_output=True, check=False)
    if res.returncode != 0:
        logging.error(res)
        return False
    return folder


def move_folder(folder, newpath, session):
    """Move or rename folder

        Args: folder - folder dict object
              newpath - string (new name/path)
              session - bytes
        Returns: Folder object on success, False on failure

    """
    fold = deepcopy(folder)
    fold['name'] = newpath
    enc = run(["bw", "encode"],
              input=json.dumps(fold).encode(),
              capture_output=True,
              check=False)
    if not enc.stdout:
        logging.error(enc)
        return False
    res = run(["bw", "edit",
               "--session", session,
               "folder", fold['id'], enc.stdout],
              capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return json.loads(res.stdout)


def add_collection(collection, org_id, session):
    """Add collection

        Args: collection - string
              org - organization id string
              session - bytes

        Returns: collection object or False on error

    """
    collection = {"name": collection, "organizationId": org_id}
    enc = run(["bw", "encode"],
              input=json.dumps(collection).encode(), capture_output=True, check=False)
    if not enc.stdout:
        logging.error(enc)
        return False
    res = run(["bw", "create",
               "--session", session,
               "--organizationid", org_id.encode(),
               "org-collection".encode(), enc.stdout],
              capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return json.loads(res.stdout)


def delete_collection(collection, session):
    """Delete collection

        Args: collection - collection object (dict)
              session - bytes
        Returns: collection object (dict) on success, False on failure

    """
    res = run(["bw", "delete",
               "--session", session,
               "--organizationid", collection['organizationId'].encode(),
               "org-collection", collection['id']],
              capture_output=True, check=False)
    if res.returncode != 0:
        logging.error(res)
        return False
    return collection


def move_collection(collection, newpath, session):
    """Move or rename collection

        Args: collection - collection object (dict)
              newpath - string (new name/path)
              session - bytes
        Returns: collection object on success, False on failure

    """
    coll = deepcopy(collection)
    coll['name'] = newpath
    enc = run(["bw", "encode"],
              input=json.dumps(coll).encode(),
              capture_output=True,
              check=False)
    if not enc.stdout:
        logging.error(enc)
        return False
    res = run(["bw", "edit",
               "--session", session,
               "--organizationid", coll['organizationId'].encode(),
               "org-collection", coll['id'], enc.stdout],
              capture_output=True, check=False)
    if not res.stdout:
        logging.error(res)
        return False
    return json.loads(res.stdout)

# vim: set et ts=4 sw=4 :



"""Editing functions for bitwarden-menu

"""
from copy import deepcopy
import os
from os.path import basename, dirname, join
import random
from secrets import choice
import shlex
import string
from subprocess import call
import tempfile

from bwm import bwcli
from bwm.bwtype import autotype_index, autotype_seq, type_text
from bwm.menu import dmenu_select, dmenu_err
import bwm


def obj_name(obj, oid):
    """Return name of folder/collection object based on id

        Args: obj - dict
              oid - string

    """
    return obj[oid]['name']


def edit_entry(entry, entries, folders, collections, session):
    # pylint: disable=too-many-branches,too-many-statements,too-many-locals
    """Edit title, username, password, url, notes and autotype sequence for an entry.

    Args: entry - selected Entry dict
          entries - list of dicts
          folders - dict of dicts {'id': {xxx,yyy}, ... }
          collections - dict of dicts {'id': {xxx,yyy}, ... }
          session - bytes
    Returns: None or entry (Item)

    """
    item = deepcopy(entry)
    update_colls = "NO"
    while True:
        fields = [str("Name: {}").format(item['name']),
                  str("Folder: {}").format(obj_name(folders, item['folderId'])),
                  str("Collections: {}").format(", ".join(obj_name(collections, i) for i
                                                          in item['collectionIds'])),
                  str("Username: {}").format(item['login']['username']),
                  str("Password: **********") if item['login']['password'] else "Password: None",
                  str("Url: {}").format(item['login']['url']),
                  str("Autotype: {}").format(autotype_seq(item)),
                  "Notes: <Enter to Edit>" if item['notes'] else "Notes: None",
                  "Delete entry",
                  "Save entry"]
        inp = "\n".join(fields)
        sel = dmenu_select(len(fields), inp=inp)
        if sel == 'Delete entry':
            delete_entry(entry, entries, session)
            return None
        if sel == 'Save entry':
            if not item.get('id'):
                res = bwcli.add_entry(item, session)
                if res is False:
                    dmenu_err("Entry not added. Check logs.")
                    return None
                entries.append(bwcli.Item(res))
            else:
                res = bwcli.edit_entry(item, session, update_colls)
                if res is False:
                    dmenu_err("Error saving entry. Changes not saved.")
                    continue
                entries[entries.index(entry)] = bwcli.Item(res)
            return bwcli.Item(res)
        try:
            field, sel = sel.split(": ", 1)
        except (ValueError, TypeError):
            return entry
        field = field.lower()
        if field == 'password':
            item = edit_password(item) or item
            continue
        if field == 'folder':
            folder = select_folder(folders)
            if folder is not False:
                item['folderId'] = folder['id']
            continue
        if field == 'collections':
            orig = item['collectionIds']
            coll_list = [collections[i] for i in collections if i in item['collectionIds']]
            collection = select_collection(collections, session, coll_list=coll_list)
            item['collectionIds'] = [*collection]
            if collection:
                item['organizationId'] = next(iter(collection.values()))['organizationId']
            if item['collectionIds'] and item['collectionIds'] != orig and orig:
                update_colls = "YES"
            elif item['collectionIds'] != orig and not orig:
                update_colls = "MOVE"
            elif not item['collectionIds'] and orig:
                update_colls = "REMOVE"
            continue
        if field == 'notes':
            item['notes'] = edit_notes(item['notes'])
            continue
        if field in ('username', 'url'):
            edit = item['login'][field] + \
                    "\n" if item['login'][field] is not None else "\n"
        elif field == 'autotype':
            edit = item['fields'][autotype_index(item)]['value'] + \
                    "\n" if item['fields'][autotype_index(item)]['value'] is not None else "\n"
        else:
            edit = item[field] + "\n" if item[field] is not None else "\n"
        sel = dmenu_select(1, f"{field.capitalize()}", inp=edit)
        if sel:
            if field in ('username', 'url'):
                item['login'][field] = sel
                if field == 'url':
                    item['login']['uris'] = [{'match': None, 'uri': sel}]
            elif field == 'autotype':
                item['fields'][autotype_index(item)]['value'] = sel
            else:
                item[field] = sel


def add_entry(entries, folders, collections, session):
    """Add vault entry

    Args: entries - list of dicts
          folders - dict of folder objects
          collections - dict of collections objects
          session - bytes
    Returns: None or entry (Item)

    """
    folder = select_folder(folders)
    colls = select_collection(collections, session, coll_list=[]) or []
    if folder is False:
        return None
    entry = {"organizationId": next(iter(colls.values()))['organizationId'] if colls else None,
             "folderId": folder['id'],
             "type": 1,
             "name": "",
             "notes": "",
             "favorite": False,
             "fields": [{"name": "autotype", "value": "", "type": 0}],
             "login": {"username": "",
                       "password": "",
                       "url": ""},
             "collectionIds": [*colls],
             "secureNote": "",
             "card": "",
             "identity": ""}
    return edit_entry(entry, entries, folders, collections, session)


def delete_entry(entry, entries, session):
    """Delete an entry

    Args: entry - dict
          entries - list of dicts
          session - bytes

    """
    inp = "NO\nYes - confirm delete\n"
    delete = dmenu_select(2, f"Confirm delete of {entry['name']}", inp=inp)
    if delete != "Yes - confirm delete":
        return
    res = bwcli.delete_entry(entry, session)
    if res is False:
        dmenu_err("Item not deleted. Check logs.")
        return
    del entries[entries.index(res)]


def edit_notes(note):
    """Use $EDITOR (or 'vim' if not set) to edit the notes entry

    In configuration file:
        Set 'gui_editor' for things like emacs, gvim, leafpad
        Set 'editor' for vim, emacs -nw, nano unless $EDITOR is defined
        Set 'terminal' if using a non-gui editor

    Args: note - string
    Returns: note - string

    """
    if bwm.CONF.has_option("vault", "gui_editor"):
        editor = bwm.CONF.get("vault", "gui_editor")
        editor = shlex.split(editor)
    else:
        if bwm.CONF.has_option("vault", "editor"):
            editor = bwm.CONF.get("vault", "editor")
        else:
            editor = os.environ.get('EDITOR', 'vim')
        if bwm.CONF.has_option("vault", "terminal"):
            terminal = bwm.CONF.get("vault", "terminal")
        else:
            terminal = "xterm"
        terminal = shlex.split(terminal)
        editor = shlex.split(editor)
        editor = terminal + ["-e"] + editor
    note = b'' if note is None else note.encode(bwm.ENC)
    with tempfile.NamedTemporaryFile(suffix=".tmp") as fname:
        fname.write(note)
        fname.flush()
        editor.append(fname.name)
        try:
            call(editor)
            fname.seek(0)
            note = fname.read()
        except FileNotFoundError:
            dmenu_err("Terminal not found. Please update config.ini.")
    note = '' if not note else note.decode(bwm.ENC)
    return note


def gen_passwd(chars, length=20):
    """Generate password (min = # of distinct character sets picked)

    Args: chars - Dict {preset_name_1: {char_set_1: string, char_set_2: string},
                        preset_name_2: ....}
          length - int (default 20)

    Returns: password - string OR False

    """
    sets = set()
    if chars:
        sets = set(j for i in chars.values() for j in i.values())
    if length < len(sets) or not chars:
        return False
    alphabet = "".join(set("".join(j for j in i.values()) for i in chars.values()))
    # Ensure minimum of one char from each character set
    password = "".join(choice(k) for k in sets)
    password += "".join(choice(alphabet) for i in range(length - len(sets)))
    tpw = list(password)
    random.shuffle(tpw)
    return "".join(tpw)


def get_password_chars():
    """Get characters to use for password generation from defaults, config file
    and user input.

    Returns: Dict {preset_name_1: {char_set_1: string, char_set_2: string},
                   preset_name_2: ....}
    """
    chars = {"upper": string.ascii_uppercase,
             "lower": string.ascii_lowercase,
             "digits": string.digits,
             "punctuation": string.punctuation}
    presets = {}
    presets["Letters+Digits+Punctuation"] = chars
    presets["Letters+Digits"] = {k: chars[k] for k in ("upper", "lower", "digits")}
    presets["Letters"] = {k: chars[k] for k in ("upper", "lower")}
    presets["Digits"] = {k: chars[k] for k in ("digits",)}
    if bwm.CONF.has_section('password_chars'):
        pw_chars = dict(bwm.CONF.items('password_chars'))
        chars.update(pw_chars)
        for key, val in pw_chars.items():
            presets[key.title()] = {k: chars[k] for k in (key,)}
    if bwm.CONF.has_section('password_char_presets'):
        if bwm.CONF.options('password_char_presets'):
            presets = {}
        for name, val in bwm.CONF.items('password_char_presets'):
            try:
                presets[name.title()] = {k: chars[k] for k in shlex.split(val)}
            except KeyError:
                dmenu_err(f"Error: Unknown value in preset {name}. Ignoring.")
                continue
    inp = "\n".join(presets)
    char_sel = dmenu_select(len(presets),
                            "Pick character set(s) to use", inp=inp)
    # This dictionary return also handles Rofi multiple select
    return {k: presets[k] for k in char_sel.split('\n')} if char_sel else False


def edit_password(entry):  # pylint: disable=too-many-return-statements
    """Edit password

        Args: entry dict
        Returns: entry dict

    """
    sel = entry['login']['password']
    pw_orig = sel + "\n" if sel is not None else "\n"
    inputs = ["Generate password",
              "Manually enter password"]
    if entry['login']['password']:
        inputs.append("Type existing password")
    pw_choice = dmenu_select(len(inputs), "Password Options", inp="\n".join(inputs))
    if pw_choice == "Manually enter password":
        sel = dmenu_select(1, "Password", inp=pw_orig)
        sel_check = dmenu_select(1, "Verify password")
        if not sel_check or sel_check != sel:
            dmenu_err("Passwords do not match. No changes made.")
            return False
    elif pw_choice == "Generate password":
        inp = "20\n"
        length = dmenu_select(1, "Password Length?", inp=inp)
        if not length:
            return False
        try:
            length = int(length)
        except ValueError:
            length = 20
        chars = get_password_chars()
        if chars is False:
            return False
        sel = gen_passwd(chars, length)
        if sel is False:
            dmenu_err("Number of char groups desired is more than requested pw length")
            return False
    elif pw_choice == "Type existing password":
        type_text(entry['login']['password'])
        return False
    else:
        return False
    entry['login']['password'] = sel
    return entry


def select_folder(folders, prompt="Folders"):
    """Select which folder for an entry

    Args: folders - dict of folder dicts ['id': {'id', 'name',...}, ...]
          options - list of menu options for folders

    Returns: False for no entry
             folder - folder object

    """
    num_align = len(str(len(folders)))
    pattern = str("{:>{na}} - {}")
    folder_names = dict(enumerate(folders.values()))
    inp = str("\n").join(pattern.format(j, i['name'], na=num_align)
                         for j, i in folder_names.items())
    sel = dmenu_select(min(bwm.MAX_LEN, len(folders)), prompt, inp=inp)
    if not sel:
        return False
    try:
        return folder_names[int(sel.split(' - ', maxsplit=1)[0])]
    except (ValueError, TypeError):
        return False


def manage_folders(folders, session):
    """Rename, create, move or delete folders

    Args: folders - dict of folder objects {'id': dict, ...}
          session - bytes

    """
    options = ['Create',
               'Move',
               'Rename',
               'Delete']
    while True:
        inp = "\n".join(i for i in options) + "\n\n" + \
            "\n".join(i['name'] for i in folders.values())
        sel = dmenu_select(len(options) + len(folders) + 1, "Manage Folders", inp=inp)
        if not sel:
            break
        if sel == 'Create':
            create_folder(folders, session)
        elif sel == 'Move':
            move_folder(folders, session)
        elif sel == 'Rename':
            rename_folder(folders, session)
        elif sel == 'Delete':
            delete_folder(folders, session)
        else:
            break


def create_folder(folders, session):
    """Create new folder

    Args: folders - dict of folder objects

    """
    parentfolder = select_folder(folders, prompt="Select parent folder")
    if parentfolder is False:
        return
    pfname = parentfolder['name']
    if pfname == "No Folder":
        pfname = ""
    name = dmenu_select(1, "Folder name")
    if not name:
        return
    name = join(pfname, name)
    folder = bwcli.add_folder(name, session)
    if folder is False:
        dmenu_err("Folder not added. Check logs.")
        return
    folders[folder['id']] = folder


def delete_folder(folders, session):
    """Delete a folder

    Args: folder - folder dict obj
          session - bytes

    """
    folder = select_folder(folders, prompt="Delete Folder:")
    if not folder or folder['name'] == "No Folder":
        return
    inp = "NO\nYes - confirm delete\n"
    delete = dmenu_select(2, "Confirm delete", inp=inp)
    if delete != "Yes - confirm delete":
        return
    res = bwcli.delete_folder(folder, session)
    if res is False:
        dmenu_err("Folder not deleted. Check logs.")
        return
    del folders[folder['id']]


def move_folder(folders, session):
    """Move folder

    Args: folders - dict {'name': folder dict, ...}

    """
    folder = select_folder(folders, prompt="Select folder to move")
    if folder is False or folder['name'] == "No Folder":
        return
    destfolder = select_folder(folders,
                               prompt="Select destination folder. 'No Folder' is root.")
    if destfolder is False:
        return
    dname = ""
    if destfolder['name'] != "No Folder":
        dname = destfolder['name']
    folder = bwcli.move_folder(folder, join(dname, basename(folder['name'])), session)
    if folder is False:
        dmenu_err("Folder not added. Check logs.")
        return
    folders[folder['id']] = folder


def rename_folder(folders, session):
    """Rename folder

    Args: folders - dict {'name': folder dict, ...}

    """
    folder = select_folder(folders, prompt="Select folder to rename")
    if folder is False or folder['name'] == "No Folder":
        return
    name = dmenu_select(1, "New folder name", inp=basename(folder['name']))
    if not name:
        return
    new = join(dirname(folder['name']), name)
    folder = bwcli.move_folder(folder, new, session)
    if folder is False:
        dmenu_err("Folder not renamed. Check logs.")
        return
    folders[folder['id']] = folder


def select_collection(collections, session,
                      prompt="Collections - Organization (ESC for no selection)",
                      coll_list=False):
    """Select which collection for an entry

    Args: collections - dict of collection dicts {'id': {'id', 'name',...}, ...}
          options - list of menu options for collections
          prompt - displayed prompt
          coll_list - list of collection objects or False if only one collection
                      will be selected

    Returns: collections - dict{id: dict, id1: dict, ...}

    """
    if coll_list is not False:
        # When multiple collections will be selected, they have to come from the
        # same organization.
        org = select_org(session)
        if org is False:
            return False
        orgs = {org['id']: org}
        prompt_name = org['name']
        prompt = f"Collections - {prompt_name} (Enter to select, ESC when done)"
        colls = {i: j for i, j in enumerate(collections.values()) if
                 j['organizationId'] == org['id']}
    else:
        orgs = bwcli.get_orgs(session)
        colls = dict(enumerate(collections.values()))
    num_align = len(str(len(colls)))
    pattern = str("{:>{na}} - {} - {}")

    def check_coll(num, coll_list, cur_coll):
        # Check if name and org_id of cur_coll match in coll_list
        # Return num if not in list, otherwise return "*num"
        for i in coll_list:
            if cur_coll['organizationId'] == i['organizationId'] and cur_coll['name'] == i['name']:
                return f"*{num}"
        return num

    loop = True
    while loop:
        if coll_list is False:
            coll_list = []
            loop = False
        inp = str("\n").join(pattern.format(check_coll(j, coll_list, i),
                                            i['name'],
                                            orgs[i['organizationId']]['name'],
                                            na=num_align)
                             for j, i in colls.items())
        sel = dmenu_select(min(bwm.MAX_LEN, len(colls)), prompt, inp=inp)
        if not sel:
            return {i['id']: i for i in coll_list}
        if sel.startswith('*'):
            sel = sel.lstrip('*')
            try:
                col = colls[int(sel.split(' - ', maxsplit=1)[0])]
                coll_list.remove(col)
            except (ValueError, TypeError):
                loop = False
        else:
            try:
                col = colls[int(sel.split(' - ', maxsplit=1)[0])]
                coll_list.append(col)
            except (ValueError, TypeError):
                loop = False
    return {i['id']: i for i in coll_list}


def manage_collections(collections, session):
    """Rename, create, move or delete collections

    Args: collections - dict of collection objects {'name': dict, ...}
          session - bytes

    """
    options = ['Create',
               'Move',
               'Rename',
               'Delete']
    while True:
        inp = "\n".join(i for i in options) + "\n\n" + \
                "\n".join(i['name'] for i in collections.values())
        sel = dmenu_select(len(options) + len(collections) + 1, "Manage collections", inp=inp)
        if not sel:
            break
        if sel == 'Create':
            create_collection(collections, session)
        elif sel == 'Move':
            move_collection(collections, session)
        elif sel == 'Rename':
            rename_collection(collections, session)
        elif sel == 'Delete':
            delete_collection(collections, session)
        else:
            break


def create_collection(collections, session):
    """Create new collection

    Args: collections - dict of collection objects

    """
    org_id = select_org(session)
    if org_id is False:
        return
    parentcollection = select_collection(collections, session,
                                         prompt="Select parent collection (Esc for no parent)")
    pname = ""
    if parentcollection:
        pname = next(iter(parentcollection.values()))['name']
    name = dmenu_select(1, "Collection name")
    if not name:
        return
    name = join(pname, name)
    collection = bwcli.add_collection(name, org_id['id'], session)
    collections[collection['id']] = collection


def delete_collection(collections, session):
    """Delete a collection

    Args: collections- dict of all collection objects
          session - bytes

    """
    collection = select_collection(collections, session, prompt="Delete collection:")
    if not collection:
        return
    collection = next(iter(collection.values()))
    inp = "NO\nYes - confirm delete\n"
    delete = dmenu_select(2, f"Confirm delete of {collection['name']}", inp=inp)
    if delete != "Yes - confirm delete":
        return
    res = bwcli.delete_collection(collection, session)
    if res is False:
        dmenu_err("Collection not deleted. Check logs.")
        return
    del collections[collection['id']]


def move_collection(collections, session):
    """Move collection

    Args: collections - dict {'name': collection dict, ...}

    """
    collection = select_collection(collections, session, prompt="Select collection to move")
    if not collection:
        return
    collection = next(iter(collection.values()))
    destcollection = select_collection(collections, session,
                                       prompt="Select destination collection "
                                              "(Esc to move to root directory)")
    if not destcollection:
        destcollection = {'name': ""}
    else:
        destcollection = next(iter(destcollection.values()))
    res = bwcli.move_collection(collection,
                                join(destcollection['name'], basename(collection['name'])),
                                session)
    if res is False:
        dmenu_err("Collection not moved. Check logs.")
        return
    collections[collection['id']] = res


def rename_collection(collections, session):
    """Rename collection

    Args: collections - dict {'name': collection dict, ...}

    """
    collection = select_collection(collections, session, prompt="Select collection to rename")
    if not collection:
        return
    collection = next(iter(collection.values()))
    name = dmenu_select(1, "New collection name", inp=basename(collection['name']))
    if not name:
        return
    new = join(dirname(collection['name']), name)
    res = bwcli.move_collection(collection, new, session)
    if res is False:
        dmenu_err("Collection not deleted. Check logs.")
        return
    collections[collection['id']] = res


def select_org(session):
    """Select organization

    Args: session - bytes

    Returns: False for no entry
             org - dict

    """
    orgs = bwcli.get_orgs(session)
    orgs_ids = dict(enumerate(orgs.values()))
    num_align = len(str(len(orgs)))
    pattern = str("{:>{na}} - {}")
    inp = str("\n").join(pattern.format(j, i['name'], na=num_align)
                         for j, i in orgs_ids.items())
    sel = dmenu_select(min(bwm.MAX_LEN, len(orgs)), "Select Organization", inp=inp)
    if not sel:
        return False
    try:
        return orgs_ids[int(sel.split(' - ', 1)[0])]
    except (ValueError, TypeError):
        return False

# vim: set et ts=4 sw=4 :



"""Bitwarden-menu main module

"""
from enum import Enum, auto
from functools import partial
import multiprocessing
import shlex
import sys
import subprocess
from threading import Timer

from bwm import bwcli
from bwm.bwedit import add_entry, edit_entry, manage_collections, manage_folders
from bwm.bwtype import type_text, type_entry
from bwm.bwview import view_all_entries, view_entry
from bwm.menu import dmenu_select, dmenu_err
import bwm


def get_passphrase(twofac=False):
    """Get a vault password from dmenu or pinentry

        Args: twofac - bool (default False) prompt for 2FA code
        Returns: string

    """
    pinentry = None
    pin_prompt = 'setdesc Enter vault password\ngetpin\n' if twofac is False \
        else 'setdesc Enter 2FA code\ngetpin\n'
    if bwm.CONF.has_option("dmenu", "pinentry"):
        pinentry = bwm.CONF.get("dmenu", "pinentry")
    if pinentry:
        password = ""
        out = subprocess.run(pinentry,
                             capture_output=True,
                             check=False,
                             encoding=bwm.ENC,
                             input=pin_prompt).stdout
        if out:
            res = out.split("\n")[2]
            if res.startswith("D "):
                password = res.split("D ")[1]
    else:
        password = dmenu_select(0, "Password" if twofac is False else "2FA Code")
    return password


def get_vault():
    # pylint: disable=too-many-return-statements,too-many-locals,too-many-branches
    """Read vault login parameters from config or ask for user input.

    Returns: Session - bytes
                       None on error opening/reading vault

    """
    args = bwm.CONF.items('vault')
    args_dict = dict(args)
    servers = [i for i in args_dict if i.startswith('server')]
    vaults = []
    for srv in servers:
        idx = srv.rsplit('_', 1)[-1]
        email = args_dict.get(f'email_{idx}', "")
        passw = args_dict.get(f'password_{idx}', "")
        twofactor = args_dict.get(f'twofactor_{idx}', None)
        if not args_dict[srv] or not email:
            continue
        try:
            cmd = args_dict[f'password_cmd_{idx}']
            res = subprocess.run(shlex.split(cmd),
                                 check=False,
                                 capture_output=True,
                                 encoding=bwm.ENC)
            if res.stderr:
                dmenu_err(f"Password command error: {res.stderr}")
                sys.exit()
            else:
                passw = res.stdout.rstrip('\n') if res.stdout else passw
        except KeyError:
            pass
        vaults.append((args_dict[srv], email, passw, twofactor))
    if not vaults or (not vaults[0][0] or not vaults[0][1]):
        res = get_initial_vault()
        if res:
            vaults.insert(0, res)
        else:
            return None
    if len(vaults) > 1:
        inp = "\n".join(i[0] for i in vaults)
        sel = dmenu_select(len(vaults), "Select Vault", inp=inp)
        vaults = [i for i in vaults if i[0] == sel]
        if not sel or not vaults:
            return None
    url, email, passw, twofactor = vaults[0]
    status = bwcli.status()
    if status is False:
        return None
    if not passw:
        passw = get_passphrase()
        if not passw:
            return None
    if status['serverUrl'] != url:
        res = bwcli.set_server(url)
        if res is False:
            return None
    if status['userEmail'] != email or status['status'] == 'unauthenticated':
        code = get_passphrase(True) if twofactor else ""
        session, err = bwcli.login(email, passw, twofactor, code)
    elif status['status'].endswith('locked'):
        session, err = bwcli.unlock(passw)
    if session is False:
        dmenu_err(err)
        return None
    return session


def get_initial_vault():
    """Ask for initial server URL and email if not entered in config file

    """
    url = dmenu_select(0, "Enter server URL.", "https://vault.bitwarden.com")
    if not url:
        dmenu_err("No URL entered. Try again.")
        return False
    email = dmenu_select(0, "Enter login email address.")
    twofa = {'None': '', 'TOTP': 0, 'Email': 1, 'Yubikey': 3}
    method = dmenu_select(len(twofa), "Select Two Factor Auth type.", "\n".join(twofa))
    with open(bwm.CONF_FILE, 'w', encoding=bwm.ENC) as conf_file:
        bwm.CONF.set('vault', 'server_1', url)
        if email:
            bwm.CONF.set('vault', 'email_1', email)
        if method:
            bwm.CONF.set('vault', 'twofactor_1', str(twofa[method]))
        bwm.CONF.write(conf_file)
    return (url, email, '', twofa[method])


def dmenu_view(entries, folders):
    """View/type individual entries (called from dmenu_run)

        Args: entries (list of dicts)
              folders (dict of dicts)

        Returns: None or entry (Item)

    """
    sel = view_all_entries([], entries, folders)
    try:
        entry = entries[int(sel.split('(', 1)[0])]
    except (ValueError, TypeError):
        return None
    text = view_entry(entry, folders)
    type_text(text)
    return entry


def dmenu_view_previous_entry(entry, folders):
    """View previous entry

        Args: entry (Item)
        Returns: entry (Item)

    """
    if entry is not None:
        text = view_entry(entry, folders)
        type_text(text)
    return entry


def dmenu_edit(entries, folders, collections, session):
    """Select items to edit (called from dmenu_run)

        Args: entries (list of dicts)
              folders (dict of dict objects)
              collections (dict of dict objects)
              session (bytes)
        Returns: None or entry (Item)

    """
    sel = view_all_entries([], entries, folders)
    try:
        entry = entries[int(sel.split('(', 1)[0])]
    except (ValueError, TypeError):
        return None
    return edit_entry(entry, entries, folders, collections, session)


def dmenu_add(entries, folders, collections, session):
    """Call add item option (called from dmenu_run)

        Args: entries (list of dicts)
              folders (dict of dict objects)
              collections (dict of dict objects)
              session (bytes)
        Returns: None or entry (Item)

    """
    return add_entry(entries, folders, collections, session)


def dmenu_folders(folders, session):
    """Call manage folders option (called from dmenu_run)

        Args: folders (dict of dict objects)
              session (bytes)
        Returns: dict {err: <Bool>, reload: <Bool>}

    """
    manage_folders(folders, session)
    return Run.CONTINUE


def dmenu_collections(collections, session):
    """Call manage collections option (called from dmenu_run)

        Args: collections (dict of dict objects)
              session (bytes)
        Returns: dict {err: <Bool>, reload: <Bool>}

    """
    manage_collections(collections, session)
    return Run.CONTINUE


def dmenu_sync(session):
    """Call vault sync option (called from dmenu_run)

        Args: session (bytes)

    """
    res = bwcli.sync(session)
    if res is False:
        dmenu_err("Sync error. Check logs.")


class Run(Enum):
    """Enum for dmenu_run return values

    """
    LOCK = auto()
    CONTINUE = auto()
    RELOAD = auto()


def dmenu_run(entries, folders, collections, session, prev_entry):
    """Run dmenu with the given list of vault Entry objects

    If 'hide_folders' is defined in config.ini, hide those from main and
    view/type all views.

    Returns: Run Enum (LOCK, CONTINUE or RELOAD)

    """
    if bwm.CONF.has_option("vault", "hide_folders"):
        hid_fold = bwm.CONF.get("vault", "hide_folders").split("\n")
        # Validate ignored folder names in config.ini
        hid_fold = [i for i in hid_fold if i in
                    [j['name'] for j in folders.values()]]
        entries_hid = [i for i in entries if i['folder'] not in hid_fold]
    else:
        entries_hid = entries
    options = {'View/Type Individual entries': partial(dmenu_view, entries_hid, folders),
               'View previous entry': partial(dmenu_view_previous_entry, prev_entry, folders),
               'Edit entries': partial(dmenu_edit, entries, folders, collections, session),
               'Add entry': partial(dmenu_add, entries, folders, collections, session),
               'Manage folders': partial(dmenu_folders, folders, session),
               'Manage collections': partial(dmenu_collections, collections, session),
               'Sync vault': partial(dmenu_sync, session),
               'Lock vault': bwcli.lock}
    sel = view_all_entries(options, entries_hid, folders)
    if not sel:
        return Run.CONTINUE
    if sel == "Lock vault":  # Kill bwm daemon
        options[sel]()
        return Run.LOCK
    if sel == "Sync vault":
        options[sel]()
        return Run.RELOAD
    if sel not in options:
        # Autotype selected entry
        try:
            entry = entries[int(sel.split('(', 1)[0])]
        except (ValueError, TypeError):
            return Run.CONTINUE
        type_entry(entry)
        return Run.CONTINUE
    return options[sel]()


class DmenuRunner(multiprocessing.Process):
    # pylint: disable=too-many-instance-attributes
    """Listen for dmenu calling event and run bwm

    Args: server - Server object

    """
    def __init__(self, server):
        multiprocessing.Process.__init__(self)
        self.server = server
        self.session = get_vault()
        if self.session is None:
            self.server.kill_flag.set()
            sys.exit()
        self.entries, self.folders, self.collections, self.orgs = bwcli.get_entries(self.session)
        self.prev_entry = None
        if not all(i for i in (self.entries, self.folders, self.collections, self.orgs)
                   if i is False):
            self.server.kill_flag.set()
            sys.exit()

    def _set_timer(self):
        """Set inactivity timer

        """
        # pylint: disable=attribute-defined-outside-init
        self.cache_timer = Timer(bwm.SESSION_TIMEOUT_MIN * 60, self.cache_time)
        self.cache_timer.daemon = True
        self.cache_timer.start()

    def run(self):
        while True:
            self.server.start_flag.wait()
            if self.server.kill_flag.is_set():
                break
            try:
                self.cache_timer.cancel()
            except AttributeError:
                pass
            self._set_timer()
            res = dmenu_run(self.entries, self.folders, self.collections,
                            self.session, self.prev_entry)
            if res == Run.LOCK:
                try:
                    self.server.kill_flag.set()
                except (EOFError, IOError):
                    return
            if res == Run.RELOAD:
                self.entries, self.folders, self.collections, self.orgs = \
                        bwcli.get_entries(self.session)
                if not all(i for i in (self.entries, self.folders, self.collections, self.orgs)
                           if i is False):
                    dmenu_err("Error loading entries. See logs.")
            if str(res) not in repr(Run.__members__):
                self.prev_entry = res or self.prev_entry
            if self.server.cache_time_expired.is_set():
                self.server.kill_flag.set()
            if self.server.kill_flag.is_set():
                break
            self.server.start_flag.clear()

    def cache_time(self):
        """Kill bwm daemon when cache timer expires

        """
        self.server.cache_time_expired.set()
        if not self.server.start_flag.is_set():
            self.server.kill_flag.set()
            self.server.start_flag.set()

# vim: set et ts=4 sw=4 :



"""Bitwarden-menu view functions

"""
from os.path import join
import webbrowser

from bwm.menu import dmenu_select
import bwm


def obj_name(obj, oid):
    """Return name of folder/collection object based on id

        Args: obj - dict
              oid - string

    """
    path = obj[oid]['name']
    if path == 'No Folder':
        path = '/'
    return path


def view_all_entries(options, vault_entries, folders):
    """Generate numbered list of all vault entries and open with dmenu.

    Returns: dmenu selection

    """
    num_align = len(str(len(vault_entries)))
    # Login: Num(l) - Folder/name - username - url
    bw_login_pattern = str("{:>{na}}(l) - {} - {} - {}")
    # Secure Note: Num(n) - Folder/name
    bw_note_pattern = str("{:>{na}}(n) - {}")
    # Card: Num(c) - Folder/name - card type - card owner - card #
    bw_card_pattern = str("{:>{na}}(c) - {} - {} - {} - {}")
    # Identity: Num(i) - Folder/name - lastName, firstName - email - phone
    bw_ident_pattern = str("{:>{na}}(i) - {} - {}, {} - {} - {}")
    # Have to number each entry to capture duplicates correctly
    ven = []
    for j, i in enumerate(vault_entries):
        if i['type'] == 1:
            ven.append(bw_login_pattern.format(
                j,
                join(obj_name(folders, i['folderId']), i['name']),
                i['login']['username'],
                i['login']['url'],
                na=num_align))
        elif i['type'] == 2:
            ven.append(bw_note_pattern.format(
                j,
                join(obj_name(folders, i['folderId']), i['name']),
                na=num_align))
        elif i['type'] == 3:
            ven.append(bw_card_pattern.format(
                j,
                join(obj_name(folders, i['folderId']), i['name']),
                i['card']['brand'],
                i['card']['cardholderName'],
                i['card']['number'],
                na=num_align))
        elif i['type'] == 4:
            ven.append(bw_ident_pattern.format(
                j,
                join(obj_name(folders, i['folderId']), i['name']),
                i['identity']['lastName'],
                i['identity']['firstName'],
                i['identity']['email'],
                i['identity']['phone'],
                na=num_align))
    vault_entries_s = str("\n").join(ven)
    if options:
        options_s = ("\n".join(options) + "\n")
        entries_s = options_s + vault_entries_s
    else:
        entries_s = vault_entries_s
    return dmenu_select(min(bwm.MAX_LEN, len(options) + len(vault_entries)), inp=entries_s)


def view_entry(entry, folders):
    """Show an entry (login, card, identity or secure note)

    """
    entry_types = {1: view_login, 2: view_note, 3: view_card, 4: view_ident}
    return entry_types[entry['type']](entry, folders)


def view_login(entry, folders):
    """Show title, username, password, url and notes for a login entry.

    Returns: dmenu selection

    """
    fields = [entry['name'] or "Title: None",
              obj_name(folders, entry['folderId']),
              entry['login']['username'] or "Username: None",
              '**********' if entry['login']['password'] else "Password: None",
              entry['login']['url'] or "URL: None",
              "Notes: <Enter to view>" if entry['notes'] else "Notes: None"]
    vault_entries = "\n".join(fields)
    sel = dmenu_select(len(fields), inp=vault_entries)
    if sel == "Notes: <Enter to view>":
        sel = view_notes(entry['notes'])
    elif sel == "Notes: None":
        sel = ""
    elif sel == '**********':
        sel = entry['login']['password']
    elif sel == fields[4]:
        if sel != "URL: None":
            webbrowser.open(sel)
        sel = ""
    return sel if not sel.endswith(": None") else ""


def view_note(entry, folders):
    """Show title and note for a secure note entry.

    Returns: dmenu selection

    """
    fields = [entry['name'] or "Title: None",
              obj_name(folders, entry['folderId']),
              "Notes: <Enter to view>" if entry['notes'] else "Notes: None"]
    vault_entries = "\n".join(fields)
    sel = dmenu_select(len(fields), inp=vault_entries)
    if sel == "Notes: <Enter to view>":
        sel = view_notes(entry['notes'])
    elif sel == "Notes: None":
        sel = ""
    return sel


def view_card(entry, folders):
    """Show title, card info and notes for a card entry.

    Returns: dmenu selection

    """
    exp = "Expiration Date: None"
    if entry['card']['expMonth'] or entry['card']['expYear']:
        exp = f"{entry['card']['expMonth']}/{entry['card']['expYear']}"
    fields = [entry['name'] or "Title: None",
              obj_name(folders, entry['folderId']),
              entry['card']['brand'] or "Card Type: None",
              entry['card']['cardholderName'] or "Card Holder Name: None",
              entry['card']['number'] or "Card Number: None",
              exp,
              entry['card']['code'] or "CVV Code: None",
              "Notes: <Enter to view>" if entry['notes'] else "Notes: None"]
    vault_entries = "\n".join(fields)
    sel = dmenu_select(len(fields), inp=vault_entries)
    if sel == "Notes: <Enter to view>":
        sel = view_notes(entry['notes'])
    elif sel == "Notes: None":
        sel = ""
    return sel


def view_ident(entry, folders):
    """Show title, identify information and notes for an identity entry.

    Returns: dmenu selection

    """
    fields = [entry['name'] or "Title: None",
              obj_name(folders, entry['folderId']),
              entry['identity']['title'] or "Title: None",
              entry['identity']['firstName'] or "First name: None",
              entry['identity']['middleName'] or "Middle name: None",
              entry['identity']['lastName'] or "Last name: None",
              entry['identity']['address1'] or "Address1: None",
              entry['identity']['address2'] or "Address2: None",
              entry['identity']['address3'] or "Address3: None",
              entry['identity']['city'] or "City: None",
              entry['identity']['state'] or "State: None",
              entry['identity']['postalCode'] or "Postal Code: None",
              entry['identity']['country'] or "Country: None",
              entry['identity']['email'] or "Email: None",
              entry['identity']['phone'] or "Phone: None",
              entry['identity']['ssn'] or "SSN: None",
              entry['identity']['username'] or "Username: None",
              entry['identity']['passportNumber'] or "Passport #: None",
              entry['identity']['licenseNumber'] or "License #: None",
              "Notes: <Enter to view>" if entry['notes'] else "Notes: None"]
    vault_entries = "\n".join(fields)
    sel = dmenu_select(len(fields), inp=vault_entries)
    if sel == "Notes: <Enter to view>":
        sel = view_notes(entry['notes'])
    elif sel == "Notes: None":
        sel = ""
    return sel


def view_notes(notes):
    """View the 'Notes' field line-by-line within dmenu.

    Returns: text of the selected line for typing

    """
    sel = dmenu_select(min(bwm.MAX_LEN, len(notes.split('\n'))), inp=notes)
    return sel



"""Dmenu/Rofi functions

"""
import shlex
import sys
from subprocess import run

import bwm


def dmenu_cmd(num_lines, prompt):
    """Parse config.ini for dmenu options

    Args: args - num_lines: number of lines to display
                 prompt: prompt to show
    Returns: command invocation (as a list of strings) for
                ["dmenu", "-l", "<num_lines>", "-p", "<prompt>", "-i", ...]

    """
    command = ["dmenu"]
    if bwm.CONF.has_option('dmenu', 'dmenu_command'):
        command = shlex.split(bwm.CONF.get('dmenu', 'dmenu_command'))
    dmenu_command = command[0]
    dmenu_args = command[1:]
    obscure = True
    obscure_color = "#222222"
    if prompt == "Password":
        if bwm.CONF.has_option('dmenu_passphrase', 'obscure'):
            obscure = bwm.CONF.getboolean('dmenu_passphrase', 'obscure')
        if bwm.CONF.has_option('dmenu_passphrase', 'obscure_color'):
            obscure_color = bwm.CONF.get('dmenu_passphrase', 'obscure_color')
    if "rofi" in dmenu_command:
        dmenu = [dmenu_command, "-dmenu", "-p", str(prompt), "-l", str(num_lines)]
        if obscure is True and prompt in ("Password", "Verify password"):
            dmenu.append("-password")
    elif "dmenu" in dmenu_command:
        dmenu = [dmenu_command, "-p", str(prompt)]
        if obscure is True and prompt in ("Password", "Verify password"):
            dmenu.extend(["-nb", obscure_color, "-nf", obscure_color])
    else:
        # Catchall for some other menu programs. Maybe it'll run and not fail?
        dmenu = [dmenu_command]
    dmenu[1:1] = dmenu_args
    return dmenu


def dmenu_select(num_lines, prompt="Entries", inp=""):
    """Call dmenu and return the selected entry

    Args: num_lines - number of lines to display
          prompt - prompt to show
          inp - string to pass to dmenu via STDIN

    Returns: sel - string

    """
    cmd = dmenu_cmd(num_lines, prompt)
    res = run(cmd,
              capture_output=True,
              check=False,
              input=inp,
              encoding=bwm.ENC,
              env=bwm.ENV)
    if res.stderr:
        cmd = [cmd[0]] + ["-dmenu"] if "rofi" in cmd[0] else [""]
        run(cmd[0], check=False, input=res.stderr, env=bwm.ENV)
        sys.exit()
    return res.stdout.rstrip('\n') if res.stdout is not None else None


def dmenu_err(prompt):
    """Pops up a dmenu prompt with an error message

    """
    return dmenu_select(1, prompt)

# vim: set et ts=4 sw=4 :



"""Set global variables. Read the config file. Create default config file if one
doesn't exist.

"""
import configparser
import locale
import logging
import os
import shlex
import sys
from os.path import exists, expanduser
from subprocess import run, DEVNULL

from bwm.menu import dmenu_err

AUTH_FILE = expanduser("~/.cache/.bwm-auth")
CONF_FILE = expanduser("~/.config/bwm/config.ini")

logging.basicConfig(filename=expanduser("~/.cache/bwm.log"), level=logging.ERROR)
LOGGER = logging.getLogger("bwm")

ENV = os.environ.copy()
ENV['LC_ALL'] = 'C'
ENC = locale.getpreferredencoding()
SESSION_TIMEOUT_DEFAULT_MIN = 360
SEQUENCE = "{USERNAME}{TAB}{PASSWORD}{ENTER}"
if not exists(CONF_FILE):
    CONF = configparser.ConfigParser()
    try:
        os.mkdir(os.path.dirname(CONF_FILE))
    except OSError:
        pass
    with open(CONF_FILE, 'w', encoding=ENC) as conf_file:
        CONF.add_section('dmenu')
        CONF.set('dmenu', 'dmenu_command', 'dmenu')
        CONF.add_section('dmenu_passphrase')
        CONF.set('dmenu_passphrase', 'obscure', 'True')
        CONF.set('dmenu_passphrase', 'obscure_color', '#222222')
        CONF.add_section('vault')
        CONF.set('vault', 'server_1', '')
        CONF.set('vault', 'email_1', '')
        CONF.set('vault', 'twofactor_1', '')
        CONF.set('vault', 'session_timeout_min ', str(SESSION_TIMEOUT_DEFAULT_MIN))
        CONF.set('vault', 'autotype_default', SEQUENCE)
        CONF.write(conf_file)
CONF = configparser.ConfigParser()
try:
    CONF.read(CONF_FILE)
except configparser.ParsingError as err:
    dmenu_err(f"Config file error: {err}")
    sys.exit()
if CONF.has_option('dmenu', 'dmenu_command'):
    command = shlex.split(CONF.get('dmenu', 'dmenu_command'))
if "-l" in command:
    MAX_LEN = int(command[command.index("-l") + 1])
else:
    MAX_LEN = 24
if CONF.has_option("vault", "session_timeout_min"):
    SESSION_TIMEOUT_MIN = int(CONF.get("vault", "session_timeout_min"))
else:
    SESSION_TIMEOUT_MIN = SESSION_TIMEOUT_DEFAULT_MIN
if CONF.has_option('vault', 'autotype_default'):
    SEQUENCE = CONF.get("vault", "autotype_default")
if CONF.has_option("vault", "type_library"):
    if CONF.get("vault", "type_library") == "xdotool":
        try:
            run(['xdotool', 'version'], check=False, stdout=DEVNULL)
        except OSError:
            dmenu_err("Xdotool not installed.\n"
                      "Please install or remove that option from config.ini")
            sys.exit()
    elif CONF.get("vault", "type_library") == "ydotool":
        try:
            run(['ydotool'], check=False, stdout=DEVNULL)
        except OSError:
            dmenu_err("Ydotool not installed.\n"
                      "Please install or remove that option from config.ini")
            sys.exit()

# vim: set et ts=4 sw=4 :
